/***********************************************************************
 * Time-stamp: <Mittwoch, 09.05.2001, 14:31:36; edited by heiko>
 * 
 * Tina's random number generators
 *
 * Monte Carlo integration test for random number generators
 *
 * Copyright by Heiko Bauke, Magdeburg 2001 
 *
 * heiko.bauke@student.uni-magdeburg.de
 *
 * This is free software. You may use and distribute this software under
 * the terms of the GNU GENERAL PUBLIC LICENSE Version 2. See COPYING
 * for details.
 *
 ***********************************************************************/


#include <cstdlib>
#include <cmath>
#include <iostream>
#include <iomanip>
#include "trng.h"

using namespace TRNG;
using namespace std;


inline double g(double x, double a) {
  return((fabs(4.0*x-2.0)+a)/(1.0+a));
}

void integral(vector<RNG *> r, long runs, long dim, long n) {
  long h, i, j, k, num_streams;
  double Is, f, Iallm, Iallsigma;
  num_streams=r.size();
  vector<double> I(num_streams);
  vector<double> Iall(runs);
  Iallm=0.0;
  for (h=0l; h<runs; ++h) {
    for (i=0l; i<num_streams; ++i)
      I[i]=0.0;
    Is=0.0;
    for (i=0l; i<n; i++) {
      for (j=0l; j<num_streams; ++j) {
	f=1.0;
	for (k=0l; k<dim; ++k) 
	  f*=g(r[j]->uniform(), (k+1l)*(k+1l))-1.0;
	I[j]+=f;
      }
    }
    for (i=0l; i<num_streams; ++i) {
      I[i]/=n;
      Is+=I[i]/num_streams;
    }
    cout << Is << endl;
    Iall[h]=Is;
    Iallm+=Iall[h];
  }
  Iallm/=static_cast<double>(runs);
  Iallsigma=0.0;
  for (i=0l; i<runs; ++i)
    Iallsigma+=(Iallm-Iall[i])*(Iallm-Iall[i]);
  Iallsigma/=static_cast<double>(runs)-1.0;
  Iallsigma/=static_cast<double>(runs);
  Iallsigma=sqrt(Iallsigma)*Student_t(0.99, runs-1l);
  cout << "mean\t: " << Iallm << endl;
  cout << "sigma\t: " << Iallsigma << endl;
  cout << "m/s\t: " << Iallm/Iallsigma << endl << endl;
}


void integral_main(RNG &r, long runs, long dim, long num_streams, long k, long n) {
  vector<RNG *> ra(num_streams);
  long i;
//    long j;
  cout << r.name() << endl;
  for (i=0l; i<num_streams; ++i) {
    ra[i]=r.copy();
//      for (j=0l; j<i; ++j)
//        ra[i]->jump(k);
  }
//    cout << "jumping" << endl;
//    integral(ra, runs, dim, n);
  for (i=0l; i<num_streams; ++i) {
    ra[i]->reset();
    ra[i]->split(num_streams, i);
  }
//    cout << "leabfrog" << endl;
  integral(ra, runs, dim, n);
  for (i=0l; i<num_streams; ++i)
    delete ra[i];
  cout << endl;
}


int main(int argc, char *argv[]) {
  if (argc!=5) {
    cerr << 
      "Tina's Random Number Generator Library\n";
    cerr << 
      "(P) & (C) by Heiko Bauke, Magdeburg 2001\n\n";
    cerr << 
      "Monte Carlo integration test V1.0\n";
    cerr << 
      "---------------------------------\n\n";
    cerr << 
      "synopsis:\n";
    cerr << 
      "$ " << argv[0] << " runs; num_streams dim n\n\n";
    cerr <<
      "usefull arguments: runs=32; num_streams=8; dim=16; n=1000000;\n\n";
    cerr << 
      "The Programm makes runs Monte Carlo integrations in dim dimensions.\n";
    cerr << 
      "Exakt integrals value is zero. num_streams parallel num_streams of random\n";
    cerr <<
      "numbers are generated by leabfrog method.\n";
  } else {
    long runs, num_streams, dim, n, k;
    runs=atol(argv[1]);
    num_streams=atol(argv[2]);
    dim=atol(argv[3]);
    n=atol(argv[4]);
    n/=num_streams;
    k=static_cast<long>(ceil(log(static_cast<double>(n)*
				 static_cast<double>(runs)*
				 static_cast<double>(dim))/log(2.0)));
    cout << dim << " dimendions" << endl;
    cout << num_streams << " num_streams" << endl;
    cout << n << " points per stream" << endl;

    ParkMiller r1;
    RAND r2;
    LCG64 r3;
    LMC2 r4;
    LMC3 r5;
    LMC4 r6;
    CLCG2 r7;
    CLCG3 r8;
    CLCG4 r9;
    EINV r10;
    EINVLCG64 r11;
    integral_main(r1, runs, dim, num_streams, k, n);
    integral_main(r2, runs, dim, num_streams, k, n);
    integral_main(r3, runs, dim, num_streams, k, n);
    integral_main(r4, runs, dim, num_streams, k, n);
    integral_main(r5, runs, dim, num_streams, k, n);
    integral_main(r6, runs, dim, num_streams, k, n);
    integral_main(r7, runs, dim, num_streams, k, n);
    integral_main(r8, runs, dim, num_streams, k, n);
    integral_main(r9, runs, dim, num_streams, k, n);
    integral_main(r10, runs, dim, num_streams, k, n);
    integral_main(r11, runs, dim, num_streams, k, n);
  }
  return(EXIT_SUCCESS);
}
