/***********************************************************************
 * Time-stamp: <Mittwoch, 09.05.2001, 14:31:36; edited by heiko>
 * 
 * Tina's random number generators
 *
 * site test for random number generators
 *
 * Copyright by Heiko Bauke, Magdeburg 2001 
 *
 * heiko.bauke@student.uni-magdeburg.de
 *
 * This is free software. You may use and distribute this software under
 * the terms of the GNU GENERAL PUBLIC LICENSE Version 2. See COPYING
 * for details.
 *
 ***********************************************************************/

#include <cstdlib>
#include <cmath>
#include <iostream>
#include <iomanip>
#include "trng.h"

using namespace TRNG;
using namespace std;


double regression(vector<double> &x, vector<double> &y, long m, long n) {
  double u, v, t1, t2, t3, t4;
  long i;
  t1=0.0;   
  t2=0.0;   
  t3=0.0;
  t4=0.0;   
  for (i=m; i<n; ++i) {
    u=log(x[i]);
    v=log(y[i]);
    t1+=u*v;
    t2+=u;     
    t3+=v;
    t4+=u*u;   
  }
  return(((n-m)*t1-t2*t3)/((n-m)*t4-t2*t2));
}


void site(vector<RNG *> r, long n, long m) {
  vector<double> S, x;
  vector<long> pos;
  long h, i, j, min, max, walkers;
  walkers=r.size();
  cout.setf(ios::fixed);
  cout << r[0]->name() << endl;
  for (i=(n-1l)%10+11; i<=n; i+=10l)
    cout << i << '\t';
  cout << endl;
  pos.resize(walkers, 0l);
  S.resize(n, 0.0);
  x.resize(n, 0.0);
  for (i=0l; i<n; ++i)
    x[i]=static_cast<double>(i+1l);
  for (h=0l; h<m; ++h) {
    min=0l;
    max=0l;
    for (j=0l; j<walkers; ++j)
      pos[j]=0l;
    for (i=0l; i<n; ++i) {
      for (j=0l; j<walkers; ++j) {
	if (r[j]->boolean())
	  ++pos[j];
	else
	  --pos[j];
	if (pos[j]<min)
	  min=pos[j];
	if (pos[j]>max)
	  max=pos[j];
      }
      S[i]+=static_cast<double>(max-min+1l)/static_cast<double>(m);
    }
  }
  for (i=(n-1l)%10+11; i<=n; i+=10l)
    cout << setprecision(9) 
	 << regression(x, S, static_cast<long>(i*0.1), i) << '\t';
  cout << endl;
}


void site_main(RNG &r, long walkers, long k, long n, long m) {
  vector<RNG *> ra(walkers);
  long i;
//    long j;
  for (i=0l; i<walkers; ++i) {
    ra[i]=r.copy();
//      for (j=0l; j<i; ++j)
//        ra[i]->jump(k);
  }
//    site(ra, n, m);
  for (i=0l; i<walkers; ++i) {
    ra[i]->reset();
    ra[i]->split(walkers, i);
  }
  site(ra, n, m);
  for (i=0l; i<walkers; ++i)
    delete ra[i];
}


int main(int argc, char *argv[]) {
  if (argc!=4) {
    cerr << 
      "Tina's Random Number Generator Library\n";
    cerr << 
      "(P) & (C) by Heiko Bauke, Magdeburg 2001\n\n";
    cerr << 
      "Site test V1.0\n";
    cerr << 
      "--------------\n\n";
    cerr << 
      "synopsis:\n";
    cerr << 
      "$ " << argv[0] << " num_walkers n m\n\n";
    cerr <<
      "usefull arguments: walkers=2; n=2500; m=1000000; m>n; n>200\n\n";
    cerr << 
      "The Programm makes at least m random walks of n steps. Random\n";
    cerr << 
      "numbers for the random walks are generated by leabfrog method.\n";
    cerr << 
      "For lage n S_t the number of sites visited after t steps has the\n";
    cerr <<
      "form S_t ~ t^(0.5). The exponent is calculated.\n\n";
    cerr << 
      "For further informations see I. Vattulainen\n";
    cerr << 
      "\"Framework for testing random numbers in parallel calculations\",\n";
    cerr <<
      "Physical Review E, Vol. 59, Number 6, pp. 7200-7204, 199\n";
  } else {
    long walkers, n, m, k;
    walkers=atol(argv[1]);
    n=atol(argv[2]);
    m=atol(argv[3]);
    if (n<200l || m<1l)
      die("invalid arguments");
    k=static_cast<long>(ceil(log(static_cast<double>(n))/log(2.0)));
    cout << walkers << " walkers" << endl;
    cout << n << " steps" << endl;
    cout << m << " walks" << endl;

    ParkMiller r1;
    RAND r2;
    LCG64 r3;
    LMC2 r4;
    LMC3 r5;
    LMC4 r6;
    CLCG2 r7;
    CLCG3 r8;
    CLCG4 r9;
    EINV r10;
    EINVLCG64 r11;
    site_main(r1, walkers, k, n, m);
    site_main(r2, walkers, k, n, m);
    site_main(r3, walkers, k, n, m);
    site_main(r4, walkers, k, n, m);
    site_main(r5, walkers, k, n, m);
    site_main(r6, walkers, k, n, m);
    site_main(r7, walkers, k, n, m);
    site_main(r8, walkers, k, n, m);
    site_main(r9, walkers, k, n, m);
    site_main(r10, walkers, k, n, m);
    site_main(r11, walkers, k, n, m);
  }
  return(EXIT_SUCCESS);
}
